#include "Proxy.hpp"
#include <sstream>
#include <iostream>
#include <windows.h>
using namespace std;

#pragma comment(lib, "ws2_32.lib")

bool StartupSockets()
{
	WSADATA wsaData;

	return NO_ERROR == WSAStartup(MAKEWORD(2,2), &wsaData);
}

void CleanupSockets()
{
	WSACleanup();
}

bool CreateSocket(int type, int protocol, SOCKET &s)
{
	s = WSASocket(AF_INET, type, protocol, 0, 0, 0);

	return s != INVALID_SOCKET;
}

bool NetBind(SOCKET s, Port port)
{
	if (s == SOCKET_ERROR)
		return false;

	// Bind socket to port
	sockaddr_in addr;
	int addr_len;

	// Fill in IPv4 sockaddr within IPv6 addr
	sockaddr_in *addr4 = reinterpret_cast<sockaddr_in*>( &addr );

	memset(addr4, 0, sizeof(*addr4));
	addr4->sin_family = AF_INET;
	addr4->sin_addr.S_un.S_addr = INADDR_ANY;
	addr4->sin_port = htons(port);

	addr_len = sizeof(sockaddr_in);

	// Attempt to bind
	return 0 == bind(s, reinterpret_cast<sockaddr*>( &addr ), addr_len);
}

Port GetBoundPort(SOCKET s)
{
    sockaddr_in6 addr;
    int namelen = sizeof(addr);

	// If socket name cannot be determined,
	if (getsockname(s, reinterpret_cast<sockaddr*>( &addr ), &namelen))
		return 0;

	// Port is placed in the same location for IPv4 and IPv6
	return ntohs(addr.sin6_port);
}


std::string SocketGetLastErrorString()
{
	return SocketGetErrorString(WSAGetLastError());
}
std::string SocketGetErrorString(int code)
{
	switch (code)
	{
    case WSAEADDRNOTAVAIL:         return "[Address not available]";
    case WSAEADDRINUSE:            return "[Address is in use]";
    case WSANOTINITIALISED:        return "[Winsock not initialized]";
    case WSAENETDOWN:              return "[Network is down]";
    case WSAEINPROGRESS:           return "[Operation in progress]";
    case WSA_NOT_ENOUGH_MEMORY:    return "[Out of memory]";
    case WSA_INVALID_HANDLE:       return "[Invalid handle]";
    case WSA_INVALID_PARAMETER:    return "[Invalid parameter]";
    case WSAEFAULT:                return "[Fault]";
    case WSAEINTR:                 return "[Interrupted]";
    case WSAEINVAL:                return "[Invalid]";
    case WSAEISCONN:               return "[Is connected]";
    case WSAENETRESET:             return "[Network reset]";
    case WSAENOTSOCK:              return "[Parameter is not a socket]";
    case WSAEOPNOTSUPP:            return "[Operation not supported]";
    case WSAESOCKTNOSUPPORT:       return "[Socket type not supported]";
    case WSAESHUTDOWN:             return "[Shutdown]";
    case WSAEWOULDBLOCK:           return "[Operation would block]";
    case WSAEMSGSIZE:              return "[Message size]";
    case WSAETIMEDOUT:             return "[Operation timed out]";
    case WSAECONNRESET:            return "[Connection reset]";
    case WSAENOTCONN:              return "[Socket not connected]";
    case WSAEDISCON:               return "[Disconnected]";
	case WSAENOBUFS:               return "[No buffer space available]";
    case ERROR_IO_PENDING:         return "[IO operation will complete in IOCP worker thread]";
    case WSA_OPERATION_ABORTED:    return "[Operation aborted]";
    case ERROR_CONNECTION_ABORTED: return "[Connection aborted locally]";
    case ERROR_NETNAME_DELETED:    return "[Socket was already closed]";
    case ERROR_PORT_UNREACHABLE:   return "[Destination port is unreachable]";
    case ERROR_MORE_DATA:          return "[More data is available]";
	}

	ostringstream oss;
	oss << "[Error code: " << code << " (0x" << hex << code << ")]";
	return oss.str();
}

IP ParseAddress(const char *ip_str)
{
	sockaddr_in addr4;
	int out_addr_len4 = sizeof(addr4);

	struct hostent *hp = gethostbyname(ip_str);
	if (!hp) return 0;

	in_addr *ip = (in_addr*)hp->h_addr_list[0];
	if (!ip) return 0;

	return ip->S_un.S_addr;
/*
	if (!WSAStringToAddressA((char*)ip_str, AF_INET, 0,
							(sockaddr*)&addr4, &out_addr_len4))
	{
		return addr4.sin_addr.S_un.S_addr;
	}

	return 0;*/
}

std::string AddressToString(IP ip)
{
	// Construct an IPv4 sockaddr, with port = 0
	sockaddr_in addr4;
	memset(&addr4, 0, sizeof(addr4));
	addr4.sin_family = AF_INET;
	addr4.sin_addr.S_un.S_addr = ip;

	// Allocate space for address string
	char addr_str4[INET_ADDRSTRLEN + 32];
	DWORD str_len4 = sizeof(addr_str4);

	// Because inet_ntop() is not supported in Windows XP, only Vista+
	if (SOCKET_ERROR == WSAAddressToStringA((sockaddr*)&addr4, sizeof(addr4),
										   0, addr_str4, &str_len4))
		return SocketGetLastErrorString();

	return addr_str4;
}

SOCKET CreateSocketWithOptions()
{
	SOCKET s;

	if (!CreateSocket(SOCK_STREAM, IPPROTO_TCP, s))
	{
		cout << "[ERROR] Unable to create proxy socket: " << SocketGetLastErrorString() << endl;
		return INVALID_SOCKET;
	}

    int buffsize = 6400000;
    if (setsockopt(s, SOL_SOCKET, SO_SNDBUF, (char*)&buffsize, sizeof(buffsize)))
	{
		cout << "[ERROR] Unable to zero the send buffer: " << SocketGetLastErrorString() << endl;
		closesocket(s);
		return INVALID_SOCKET;
	}

    int rcv_buffsize = 6400000;
	if (setsockopt(s, SOL_SOCKET, SO_RCVBUF, (char*)&rcv_buffsize, sizeof(rcv_buffsize)))
	{
		cout << "[ERROR] Unable to zero the recv buffer: " << SocketGetLastErrorString() << endl;
		closesocket(s);
		return INVALID_SOCKET;
	}

    BOOL nodelay = TRUE;
	if (setsockopt(s, IPPROTO_TCP, TCP_NODELAY, (char*)&nodelay, sizeof(nodelay)))
	{
		cout << "[ERROR] Unable to disable Nagle algorithm: " << SocketGetLastErrorString() << endl;
		closesocket(s);
		return INVALID_SOCKET;
	}

	return s;
}

void TCPMultiplexer::WriteLine(unsigned int file_pos, unsigned char *line, int line_bytes, bool c2s)
{
	static const char HEX_TABLE[16] = { '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', 'a', 'b', 'c', 'd', 'e', 'f' };

	static const int LINE_SIZE = 16;
	static const int OUTPUT_LINE_SIZE = 8+2+LINE_SIZE*3+1+LINE_SIZE+2;

	char out_buffer[OUTPUT_LINE_SIZE];

	out_buffer[0] = c2s ? 'C' : 'S';
	out_buffer[1] = '2';
	out_buffer[2] = c2s ? 'S' : 'C';
	out_buffer[3] = ' ';
	out_buffer[4] = HEX_TABLE[(file_pos >> 12) & 15];
	out_buffer[5] = HEX_TABLE[(file_pos >> 8) & 15];
	out_buffer[6] = HEX_TABLE[(file_pos >> 4) & 15];
	out_buffer[7] = HEX_TABLE[file_pos & 15];
	out_buffer[8] = ' ';
	out_buffer[9] = ' ';

	char *hex_buffer = &out_buffer[8+2];
	int line_pos;

	for (line_pos = 0; line_pos < line_bytes; ++line_pos)
	{
		unsigned char byte = line[line_pos];

		hex_buffer[line_pos*3+0] = HEX_TABLE[byte >> 4];
		hex_buffer[line_pos*3+1] = HEX_TABLE[byte & 15];
		hex_buffer[line_pos*3+2] = ' ';
	}

	for (; line_pos < LINE_SIZE; ++line_pos)
	{
		hex_buffer[line_pos*3+0] = ' ';
		hex_buffer[line_pos*3+1] = ' ';
		hex_buffer[line_pos*3+2] = ' ';
	}

	char *ascii_buffer = &hex_buffer[LINE_SIZE*3];
	int ascii_pos;

	*ascii_buffer++ = ' ';

	for (ascii_pos = 0; ascii_pos < line_bytes; ++ascii_pos)
	{
		char ch = (char)line[ascii_pos];

		if (ch >= ' ' && ch <= '~')
		{
			ascii_buffer[ascii_pos] = ch;
		}
		else
		{
			ascii_buffer[ascii_pos] = '.';
		}
	}

	ascii_buffer[ascii_pos++] = '\r';
	ascii_buffer[ascii_pos++] = '\n';

	int ascii_bytes = &ascii_buffer[ascii_pos] - out_buffer;

	_os.write(out_buffer, ascii_bytes);
	_os.flush();
}

void TCPMultiplexer::WriteLog(unsigned char *buffer, int bytes, bool c2s)
{
	cout << (c2s ? "C --> S" : "S --> C") << "  Captured " << bytes << endl;

	unsigned int offset = 0;

	while (bytes >= 16)
	{
		WriteLine(offset, buffer, 16, c2s);
		buffer += 16;
		bytes -= 16;
		offset += 16;
	}

	if (bytes > 0)
	{
		WriteLine(offset, buffer, bytes, c2s);
	}
}


Socket::Socket()
{
	_multiplexer = 0;
	_s = INVALID_SOCKET;
	_h = INVALID_HANDLE_VALUE;

	_data_queue = 0;
	_data_queue_size = 0;
	_ready = false;
}

Socket::~Socket()
{
	if (_s != INVALID_SOCKET)
	{
		closesocket(_s);
		_s = INVALID_SOCKET;
	}

	if (_h != INVALID_HANDLE_VALUE)
	{
		WSACloseEvent(_h);
		_h = INVALID_HANDLE_VALUE;
	}

	if (_data_queue)
	{
		delete []_data_queue;
		_data_queue = 0;
		_data_queue_size = 0;
	}
}

void Socket::Initialize(TCPMultiplexer *multiplexer, HANDLE h, SOCKET s)
{
	_multiplexer = multiplexer;
	_h = h;
	_s = s;
}

bool Socket::Post(char *data, int bytes)
{
	if (!_ready)
	{
		char *nb = new char[_data_queue_size + bytes];
		if (!nb) return false;
		if (_data_queue)
		{
			memcpy(nb, _data_queue, _data_queue_size);
			delete []_data_queue;
		}

		memcpy(nb + _data_queue_size, data, bytes);
		_data_queue_size += bytes;
		_data_queue = nb;

		return true;
	}

	WSABUF wsabuf;
	wsabuf.buf = data;
	wsabuf.len = bytes;

	DWORD sent = 0;
	if (SOCKET_ERROR == WSASend(_s, &wsabuf, 1, &sent, 0, 0, 0) &&
		WSAGetLastError() != WSAEWOULDBLOCK)
	{
		return false;
	}

	return true;
}

void Socket::FlushQueued()
{
	_ready = true;

	if (_data_queue)
	{
		cout << "-- Posting " << _data_queue_size << " queued bytes now." << endl;

		Post(_data_queue, _data_queue_size);

		delete []_data_queue;
		_data_queue = 0;
		_data_queue_size = 0;
	}
}


FTPParser::FTPParser()
{
	_buffer = 0;
	_buffer_size = 0;
}

FTPParser::~FTPParser()
{
	if (_buffer)
	{
		delete []_buffer;
		_buffer = 0;
		_buffer_size = 0;
	}
}

void FTPParser::AddBuffer(char *buffer, int bytes)
{
	char *nb = new char[_buffer_size + bytes];

	if (!nb)
	{
		cout << "[ERROR] Unable to allocate memory for receive buffer " << bytes << endl << endl;
		return;
	}

	if (_buffer)
	{
		memcpy(nb, _buffer, _buffer_size);
		delete []_buffer;
	}
	_buffer = nb;

	memcpy(nb + _buffer_size, buffer, bytes);
	_buffer_size += bytes;

	bool found_line;
	do 
	{
		found_line = false;

		for (int ii = 0; ii < _buffer_size; ++ii)
		{
			char ch = _buffer[ii];

			if (ch == '\r' || ch == '\n')
			{
				if (ii > 0) OnLine(_buffer, ii);

				memmove(_buffer, _buffer + ii + 1, _buffer_size - ii - 1);
				_buffer_size = _buffer_size - ii - 1;

				found_line = true;
				break;
			}
		}
	} while (found_line);
}


ControlServer::ControlServer(IP target_ip, Port target_port)
{
	_target_ip = target_ip;
	_target_port = target_port;
	FlushQueued();
}

void ControlServer::OnAccept(SOCKET s, sockaddr_in *acceptAddr)
{
	HANDLE h = WSACreateEvent();

	if (h == INVALID_HANDLE_VALUE)
	{
		cout << "[ERROR] Unable to create control connexion event: " << SocketGetLastErrorString() << endl << endl;
		closesocket(s);
		return;
	}

	if (SOCKET_ERROR == WSAEventSelect(s, h, FD_READ|FD_WRITE|FD_CLOSE))
	{
		cout << "[ERROR] Unable to event select control connexion object: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	ControlConnexion *connexion = new ControlConnexion(*acceptAddr, _target_ip, _target_port);
	Socket *obj = connexion;
	if (!obj)
	{
		cout << "[ERROR] Unable to allocate memory for control connexion object" << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	obj->Initialize(_multiplexer, h, s);
	_multiplexer->AddSocket(obj);

	cout << "<ControlServer> Accepted connexion from " << AddressToString(acceptAddr->sin_addr.S_un.S_addr) << ":" << ntohs(acceptAddr->sin_port) << endl;

	// Client:

	s = CreateSocketWithOptions();
	if (INVALID_SOCKET == s)
	{
		cout << "[ERROR] Unable to create control client socket: " << SocketGetLastErrorString() << endl << endl;
		return;
	}

	h = WSACreateEvent();
	if (h == INVALID_HANDLE_VALUE)
	{
		cout << "[ERROR] Unable to create control client event: " << SocketGetLastErrorString() << endl << endl;
		return;
	}

	if (SOCKET_ERROR == WSAEventSelect(s, h, FD_CONNECT|FD_READ|FD_WRITE|FD_CLOSE))
	{
		cout << "[ERROR] Unable to event select control client object: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return;
	}

	// Do not allow other applications to bind over us with SO_REUSEADDR
	int exclusive = TRUE;
	if (setsockopt(s, SOL_SOCKET, SO_EXCLUSIVEADDRUSE, (char*)&exclusive, sizeof(exclusive)))
	{
		cout << "[ERROR] Unable to bind exclusive control client socket: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return;
	}

	if (!NetBind(s, 0))
	{
		cout << "[ERROR] Unable to bind control client socket: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return;
	}

	sockaddr_in addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(_target_port);
	addr.sin_addr.S_un.S_addr = _target_ip;

	if (connect(s, (sockaddr*)&addr, sizeof(addr)) && WSAGetLastError() != WSAEWOULDBLOCK)
	{
		cout << "[ERROR] Unable to connect with control client: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return;
	}

	ControlClient *client = new ControlClient(connexion);
	obj = client;
	if (!obj)
	{
		cout << "[ERROR] Unable to allocate memory for control client object" << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return;
	}

	obj->Initialize(_multiplexer, h, s);
	_multiplexer->AddSocket(obj);

	cout << "<ControlServer> Connecting to " << AddressToString(_target_ip) << ":" << _target_port << endl;

	connexion->_client = client;
}

void ControlServer::OnClose()
{
	cout << "<ControlServer> Closed" << endl;
	_multiplexer->RemoveSocket(this);
}


ControlConnexion::ControlConnexion(sockaddr_in &addr, IP target_ip, Port target_port)
{
	_addr = addr;
	_target_ip = target_ip;
	_target_port = target_port;
	_client = 0;
	FlushQueued();
}

void ControlConnexion::OnRead(char *data, int bytes)
{
	_multiplexer->WriteLog((unsigned char*)data, bytes, true);

	AddBuffer(data, bytes);
}

Port ControlConnexion::CreateDataServer(Port desired_port, Port client_port)
{
	SOCKET s = CreateSocketWithOptions();

	if (s == INVALID_SOCKET)
	{
		cout << "[ERROR] Unable to create data server socket: " << SocketGetLastErrorString() << endl << endl;
		return 0;
	}

	HANDLE h = WSACreateEvent();

	if (h == INVALID_HANDLE_VALUE)
	{
		cout << "[ERROR] Unable to create data server event: " << SocketGetLastErrorString() << endl << endl;
		closesocket(s);
		return 0;
	}

	if (SOCKET_ERROR == WSAEventSelect(s, h, FD_ACCEPT|FD_CLOSE))
	{
		cout << "[ERROR] Unable to event select data server object: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return 0;
	}

	if (!NetBind(s, 0))
	{
		cout << "[ERROR] Unable to bind data server socket: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return 0;
	}

	if (listen(s, SOMAXCONN))
	{
		cout << "[ERROR] Unable to listen on data server socket: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return 0;
	}

	Socket *obj = new DataServer(_addr.sin_addr.S_un.S_addr, client_port);
	if (!obj)
	{
		cout << "[ERROR] Unable to allocate memory for data server object" << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return 0;
	}

	obj->Initialize(_multiplexer, h, s);
	_multiplexer->AddSocket(obj);

	return GetBoundPort(s);
}

void ControlConnexion::OnLine(char *line, int bytes)
{
	static char TERMINATOR[2] = { '\r', '\n' };

	if (bytes >= 5 && !strncmp(line, "PORT ", 5))
	{
		int octets[4], port_high = 0, port_low = 0;

		std::string s(line, bytes);

		Port desired_port = GetBoundPort(_client->GetSocket()) + 1;

		if (sscanf(s.c_str(), "PORT %d,%d,%d,%d,%d,%d", &octets[0], &octets[1], &octets[2], &octets[3], &port_high, &port_low) == 6)
		{
			int port = port_high * 256 + port_low;

			Port proxy_port = CreateDataServer(desired_port, port);

			if (proxy_port)
			{
				cout << "<ControlConnexion> Captured a PORT command (active FTP mode).  Successfully created a data server to handle this request." << endl;

				cout << "<ControlConnexion> Injecting hacked PORT command to route proxy data port " << proxy_port << " to local data port " << port << endl;

				sockaddr_in myaddr;
				int addrlen = sizeof(myaddr);
				getsockname(_client->GetSocket(), (sockaddr*)&myaddr, &addrlen);

				IP ip = myaddr.sin_addr.S_un.S_addr;

				ostringstream oss;
				oss << "PORT " << (ip & 255) << "," << ((ip >> 8) & 255) << "," << ((ip >> 16) & 255) << "," << (ip >> 24) << "," << (proxy_port >> 8) << "," << (proxy_port & 255) << "\r\n";
				_client->Post((char*)oss.str().c_str(), oss.str().size());
			}
			else
			{
				cout << "[ERROR] <ControlConnexion> Captured a PORT command (active FTP mode).  Unable to create a data server!" << endl;
			}
		}
		else
		{
			cout << "[ERROR] <ControlConnexion> Captured a PORT command (active FTP mode).  Unable to parse it!" << endl;
		}

		return;
	}

	cout << "Client --> Proxy [Control] ";
	cout.write(line, bytes);
	cout << endl;

	if (!_client->Post(line, bytes) ||
		!_client->Post(TERMINATOR, 2))
	{
		cout << "Failure to relay message to target!" << endl;
		OnClose();
	}
}

void ControlConnexion::OnClose()
{
	if (_client)
	{
		ControlClient *client = _client;
		_client = 0;
		client->OnClose();

		cout << "<ControlConnexion> Client disconnected" << endl;
		_multiplexer->RemoveSocket(this);
	}
}


ControlClient::ControlClient(ControlConnexion *connexion)
{
	_connexion = connexion;
}

void ControlClient::OnConnect()
{
	cout << "<ControlClient> Connected to " << AddressToString(_connexion->_target_ip) << ":" << _connexion->_target_port << " using local port " << GetBoundPort(_s) << ". Flushing queue..." << endl;

	FlushQueued();
}

void ControlClient::OnRead(char *data, int bytes)
{
	_multiplexer->WriteLog((unsigned char*)data, bytes, false);

	AddBuffer(data, bytes);
}

void ControlClient::OnLine(char *line, int bytes)
{
	static char TERMINATOR[2] = { '\r', '\n' };

	cout << "Proxy <-- Server [Control] ";
	cout.write(line, bytes);
	cout << endl;

	if (!_connexion->Post(line, bytes) ||
		!_connexion->Post(TERMINATOR, 2))
	{
		OnClose();
	}
}

void ControlClient::OnClose()
{
	if (_connexion)
	{
		ControlConnexion *connexion = _connexion;
		_connexion = 0;
		connexion->OnClose();

		cout << "<ControlClient> Disconnected" << endl;
		_multiplexer->RemoveSocket(this);
	}
}


DataServer::DataServer(IP target_ip, Port target_port)
{
	_target_ip = target_ip;
	_target_port = target_port;
	FlushQueued();
}

void DataServer::OnAccept(SOCKET s, sockaddr_in *acceptAddr)
{
	HANDLE h = WSACreateEvent();

	if (h == INVALID_HANDLE_VALUE)
	{
		cout << "[ERROR] Unable to create data connexion event: " << SocketGetLastErrorString() << endl << endl;
		closesocket(s);
		return;
	}

	if (SOCKET_ERROR == WSAEventSelect(s, h, FD_READ|FD_WRITE|FD_CLOSE))
	{
		cout << "[ERROR] Unable to event select data connexion object: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	DataConnexion *connexion = new DataConnexion(*acceptAddr, _target_ip, _target_port);
	Socket *obj = connexion;
	if (!obj)
	{
		cout << "[ERROR] Unable to allocate memory for data connexion object" << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	obj->Initialize(_multiplexer, h, s);
	_multiplexer->AddSocket(obj);

	cout << "<DataServer> Accepted data connexion from " << AddressToString(acceptAddr->sin_addr.S_un.S_addr) << ":" << ntohs(acceptAddr->sin_port) << endl;

	// Client:

	s = CreateSocketWithOptions();
	if (INVALID_SOCKET == s)
	{
		cout << "[ERROR] Unable to create data client socket: " << SocketGetLastErrorString() << endl << endl;
		return;
	}

	h = WSACreateEvent();
	if (h == INVALID_HANDLE_VALUE)
	{
		cout << "[ERROR] Unable to create data client event: " << SocketGetLastErrorString() << endl << endl;
		return;
	}

	if (SOCKET_ERROR == WSAEventSelect(s, h, FD_CONNECT|FD_READ|FD_WRITE|FD_CLOSE))
	{
		cout << "[ERROR] Unable to event select data client object: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	sockaddr_in addr;
	memset(&addr, 0, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_port = htons(_target_port);
	addr.sin_addr.S_un.S_addr = _target_ip;

	if (connect(s, (sockaddr*)&addr, sizeof(addr)) && WSAGetLastError() != WSAEWOULDBLOCK)
	{
		cout << "[ERROR] Unable to connect with data client: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	DataClient *client = new DataClient(connexion);
	obj = client;
	if (!obj)
	{
		cout << "[ERROR] Unable to allocate memory for data client object" << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
	}

	obj->Initialize(_multiplexer, h, s);
	_multiplexer->AddSocket(obj);

	cout << "<DataServer> Connecting to " << AddressToString(_target_ip) << ":" << _target_port << endl;

	connexion->_client = client;
}

void DataServer::OnClose()
{
	cout << "<DataServer> Closed" << endl;
	_multiplexer->RemoveSocket(this);
}


DataConnexion::DataConnexion(sockaddr_in &addr, IP target_ip, Port target_port)
{
	_addr = addr;
	_target_ip = target_ip;
	_target_port = target_port;
	FlushQueued();
}

void DataConnexion::OnRead(char *data, int bytes)
{
	cout << "<DataConnexion> Captured " << bytes << " bytes from data downlink!" << endl;

	_multiplexer->WriteLog((unsigned char*)data, bytes, false);

	_client->Post(data, bytes);
}

void DataConnexion::OnClose()
{
	if (_client)
	{
		DataClient *client = _client;
		_client = 0;

		if (client->IsReady())
		{
			client->OnClose();
		}
		else
		{
			client->_connexion = 0; // unlink self
		}

		cout << "<DataConnexion> Target terminated data downlink" << endl;
		_multiplexer->RemoveSocket(this);
	}
}


DataClient::DataClient(DataConnexion *connexion)
{
	_connexion = connexion;
}

void DataClient::OnConnect()
{
	cout << "<DataClient> Connected.  Flushing queue..." << endl;
	FlushQueued();

	if (!_connexion)
	{
		cout << "<DataData> Corresponding downlink through connexion was already closed so closing after posting queued data." << endl;

		_multiplexer->RemoveSocket(this);
	}
}

void DataClient::OnRead(char *data, int bytes)
{
	cout << "<DataClient> Captured " << bytes << " bytes from data uplink!" << endl;

	_multiplexer->WriteLog((unsigned char*)data, bytes, true);

	if (_connexion) _connexion->Post(data, bytes);
}

void DataClient::OnClose()
{
	if (_connexion)
	{
		DataConnexion *connexion = _connexion;
		_connexion = 0;
		connexion->OnClose();

		cout << "<DataClient> Client terminated data uplink" << endl;
		_multiplexer->RemoveSocket(this);
	}
}


bool TCPMultiplexer::AddSocket(Socket *s)
{
	_handles.push_back(s->GetHandle());
	_sockets.push_back(s);

	return true;
}

void TCPMultiplexer::RemoveSocket(Socket *s)
{
	for (std::vector<Socket*>::iterator ii = _sockets.begin(); ii != _sockets.end(); ++ii)
	{
		if (*ii == s)
		{
			HANDLE h = s->GetHandle();

			for (std::vector<HANDLE>::iterator jj = _handles.begin(); jj != _handles.end(); ++jj)
			{
				if (*jj == h)
				{
					_handles.erase(jj);
					break;
				}
			}

			_sockets.erase(ii);
			break;
		}
	}

	delete s;
}

void TCPMultiplexer::Shutdown()
{
	for (int ii = 0, size = (int)_sockets.size(); ii < size; ++ii)
	{
		delete _sockets[ii];
	}

	_sockets.clear();
	_handles.clear();
}

TCPMultiplexer::TCPMultiplexer(ostream &os)
	: _os(os)
{
}

bool TCPMultiplexer::Initialize(Port proxy_port, IP target_ip, Port target_port)
{
	SOCKET s = CreateSocketWithOptions();

	if (s == INVALID_SOCKET)
	{
		cout << "[ERROR] Unable to create control server socket: " << SocketGetLastErrorString() << endl << endl;
		return false;
	}

	HANDLE h = WSACreateEvent();

	if (h == INVALID_HANDLE_VALUE)
	{
		cout << "[ERROR] Unable to create control server event: " << SocketGetLastErrorString() << endl << endl;
		closesocket(s);
		return false;
	}

	if (SOCKET_ERROR == WSAEventSelect(s, h, FD_ACCEPT|FD_CLOSE))
	{
		cout << "[ERROR] Unable to event select control server object: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return false;
	}

	if (!NetBind(s, proxy_port))
	{
		cout << "[ERROR] Unable to bind control server socket: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return false;
	}

	if (listen(s, SOMAXCONN))
	{
		cout << "[ERROR] Unable to listen on control server socket: " << SocketGetLastErrorString() << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return false;
	}

	Socket *obj = new ControlServer(target_ip, target_port);
	if (!obj)
	{
		cout << "[ERROR] Unable to allocate memory for control server object" << endl << endl;
		WSACloseEvent(h);
		closesocket(s);
		return false;
	}

	obj->Initialize(this, h, s);
	AddSocket(obj);

	return true;
}

TCPMultiplexer::~TCPMultiplexer()
{
}

void TCPMultiplexer::Spin()
{
	for (;;)
	{
		DWORD wait_result = WSAWaitForMultipleEvents(_handles.size(), &_handles[0], FALSE, WSA_INFINITE, FALSE);
		int event_index = wait_result - WSA_WAIT_EVENT_0;

		if (event_index < 0 || event_index >= (int)_handles.size())
		{
			cout << "Wait failure: " << SocketGetLastErrorString() << endl << endl;
			return;
		}

		Socket *s = _sockets[event_index];
		WSANETWORKEVENTS net_events;

		if (SOCKET_ERROR == WSAEnumNetworkEvents(s->GetSocket(), s->GetHandle(), &net_events))
		{
			cout << "Unable to enum events: " << SocketGetLastErrorString() << endl << endl;
			return;
		}

		if ((net_events.lNetworkEvents & FD_ACCEPT) != 0)
		{
			if (net_events.iErrorCode[FD_ACCEPT_BIT])
			{
				s->OnClose();
				continue;
			}
			else
			{
				sockaddr_in addr;
				int addrlen = sizeof(addr);
				SOCKET acceptSocket = accept(s->GetSocket(), (sockaddr*)&addr, &addrlen);

				if (acceptSocket != INVALID_SOCKET)
				{
					s->OnAccept(acceptSocket, &addr);
				}
			}
		}

		if ((net_events.lNetworkEvents & FD_CONNECT) != 0)
		{
			if (net_events.iErrorCode[FD_CONNECT_BIT])
			{
				s->OnClose();
				continue;
			}
			else
			{
				s->OnConnect();
			}
		}

		if ((net_events.lNetworkEvents & FD_READ) != 0)
		{
			if (net_events.iErrorCode[FD_READ_BIT])
			{
				s->OnClose();
				continue;
			}
			else
			{
				static char buffer[6553600];

				for (;;)
				{
					int bytes = recv(s->GetSocket(), buffer, sizeof(buffer), 0);

					if (bytes <= 0)
					{
						if (WSAGetLastError() == WSAEWOULDBLOCK)
							break;

						s->OnClose();
						break;
					}
					else
					{
						s->OnRead(buffer, bytes);
					}
				}

				continue;
			}
		}

		if ((net_events.lNetworkEvents & FD_CLOSE) != 0)
		{
			s->OnClose();
			continue;
		}
	}
}
