#include <iostream>
#include <fstream>
#include "Proxy.hpp"
using namespace std;

int main(int argc, char **argv)
{
	if (argc != 5)
	{
		if (argc >= 1)
		{
			cout << "Syntax: \"" << argv[0] << "\" <Proxy Port> <Target IP> <Target Port> <Log File Path>" << endl << endl;
		}
		return 1;
	}

	if (!StartupSockets())
	{
		cout << "[ERROR] Unable to initialize sockets." << endl << endl;
		return 2;
	}

	Port proxy_port, target_port;
	IP target_ip;
	bool input_failure = false;

	proxy_port = atoi(argv[1]);
	target_ip = ParseAddress(argv[2]);
	target_port = atoi(argv[3]);

	if (proxy_port <= 0 || proxy_port > 65535)
	{
		cout << "[ERROR] Invalid PROXY PORT specified." << endl;
		input_failure = true;
	}

	if (!target_ip)
	{
		cout << "[ERROR] Invalid TARGET IP specified: " << SocketGetLastErrorString() << endl;
		input_failure = true;
	}

	if (target_port <= 0 || target_port > 65535)
	{
		cout << "[ERROR] Invalid TARGET PORT specified." << endl;
		input_failure = true;
	}

	ofstream file(argv[4]);

	if (!file)
	{
		cout << "[ERROR] Unable to open output log file." << endl;
		input_failure = true;
	}

	if (input_failure)
	{
		cout << endl;
	}
	else
	{
		TCPMultiplexer multiplexer(file);

		if (!multiplexer.Initialize(proxy_port, target_ip, target_port))
		{
			cout << "[ERROR] Unable to initialize proxy server." << endl << endl;
		}
		else
		{
			cout << "Proxy is waiting for a connection on port " << proxy_port << "..." << endl << endl;

			multiplexer.Spin();
		}
	}

	CleanupSockets();

	return 0;
}
